/*---------------------------------------------------------------------------*\
 *           interactive networked Virtual Reality system (inVRs)            *
 *                                                                           *
 *    Copyright (C) 2005-2009 by the Johannes Kepler University, Linz        *
 *                                                                           *
 *                            www.inVRs.org                                  *
 *                                                                           *
 *              contact: canthes@inVRs.org, rlander@inVRs.org                *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                        Project: CAVESceneManager                          *
 *                                                                           *
 * The CAVE Scene Manager (CSM) is a wrapper for the OpenSG multi-display    *
 * functionality, which allows the easy configuration of a multi-display     *
 * setup via text files. It was developed by Adrian Haffegee of the Centre   *
 * for Advanced Computing and Emerging Technologies (ACET) at the            *
 * University of Reading, UK as part of his MSc Thesis.                      *
 *                                                                           *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

/*! \file OSGCAVESceneManager.cpp
 *
 *  Adrian Haffegee, Sept 2004
 *	Implementation file for CAVE Scene Manager (CSM) - based on OpenSG SSM
 */

//---------------------------------------------------------------------------
//  Includes
//---------------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>

#include <OpenSG/OSGConfig.h>
#include <OpenSG/OSGBaseFunctions.h>
#include <OpenSG/OSGSimpleGeometry.h>
#include <OpenSG/OSGImageFileType.h>
#include <OpenSG/OSGSolidBackground.h>
#include <OpenSG/OSGViewport.h>
#include <OpenSG/OSGImageBackground.h>
#include <OpenSG/OSGLogoData.h>
#include <OpenSG/OSGSimpleStatisticsForeground.h>
#include <OpenSG/OSGStatElemTypes.h>
#include <OpenSG/OSGStatCollector.h>
#include <OpenSG/OSGDrawable.h>
#include <OpenSG/OSGPointLight.h>
#include <OpenSG/OSGSpotLight.h>
#include <OpenSG/OSGDirectionalLight.h>

#include "../include/OpenSG/OSGCAVESceneManager.h"

#if defined(OSG_WIN32_ICL) && !defined(OSG_CHECK_FIELDSETARG)
#pragma warning (disable : 383)
#endif

//typedef CoredNodePtr<ComponentTransform> CTNodePtr;

OSG_USING_NAMESPACE

/***************************************************************************\
 *                            Description                                  *
\***************************************************************************/

/*! \class osg::CAVESceneManager

    CAVESceneManager manages the OpenSG components that are needed to do
    multiscreen mono/stereo applications that want to display some geometry.
	It takes the multiwindow to render into and the root node of the graph
	to display and takes care of creating the necessary additional OpenSG
	objects, like viewports camers etc. Need a CSM config file!
 */


/*! \var CAVESceneManager::_win
    The managed window. Has to be set by the user.
 */

/*! \var CAVESceneManager::_root
    The root node to be rendered. Has to be set by the user.
 */

/*! \var CAVESceneManager::_highlight
    The Node to be highlighted (highlit? ;).  Has to be set by the user.
    May be NullFC (which is the default), in which case nothing is highlighted.
 */

/*! \var CAVESceneManager::_highlightNode
    The node used to hold the actual highlight geometry.
 */

/*! \var CAVESceneManager::_highlightPoints
    The highlight points geo property. These are adjusted to include the
    _highlight object for every rendered frame.
 */

/*! \var CAVESceneManager::_internalRoot
    The real root used internally. It's the root for the light and has the
    camera beacon and the given root as children.
 */

/*! \var CAVESceneManager::_headlight
    The core for the light source.
 */

/*! \var CAVESceneManager::_foreground
    ImageForeground used by the logo.
 */

/*! \var CAVESceneManager::_statstate
    Bool to indicate if statistics should be displayed or not.
 */

/*! \var CAVESceneManager::_action
    The action used to render the scene.
 */

/*! \var CAVESceneManager::_camTrolleyTrans
    The core of the camera beacons. Based on user's head
 */

/*! \var CAVESceneManager::_cameras
    The cameras used to view the scene.
 */

/*! \var CAVESceneManager::_mousebuttons
    Storage for the actively pressed mouse buttons. Needed to do the right
    thing for mouse motions, as mouseMove() doesn't get them.
 */


/*! \var CAVESceneManager::_highlightMaterial
    The material used by the highlight object.
 */
SimpleMaterialPtr CAVESceneManager::_highlightMaterial;

StatElemDesc<StatRealElem  > amhStatDesc ("amhStat","amhStat desc");


/*-------------------------------------------------------------------------*/
/*                            Constructors                                 */

/*! Constructor
 */

CAVESceneManager::CAVESceneManager(void) :
_win            (NullFC),
	_root           (NullFC),
	_wand           (NullFC),
	_foreground     (NullFC),
	_statforeground (NullFC),
	_statstate      (false),

	_highlight      (NullFC),
	_highlightNode  (NullFC),
	_highlightPoints(NullFC),

    _internalRoot   (NullFC),
    _headlight      (NullFC),
    _litScene       (NullFC),
    _action         (NULL  ),
//    _cart           (NullFC),
    _camTrolleyTrans (NullFC),
//    _camera         (NullFC),
//    _navigator      (      ),

	_lastx          (TypeTraits<Int16>::getMax()),
	_lasty          (TypeTraits<Int16>::getMax()),
	_mousebuttons   (0                             ),
//    _stereo         (false),
	_userTransform  (NullFC),
	_yRotation (0),
	_useShadows( false )
{
	_cfg=NULL;
	for (int wallCnt=0; wallCnt<_cfg->getNumActiveWalls(); wallCnt++) {
		//    for (int camCnt=FirstCamera; camCnt<LastCamera; camCnt++) {
		_cameras[wallCnt] = NullFC;
	}

	/* ADDED BY LANDI */
	_scaleFactor = 1;
}


	CAVESceneManager::CAVESceneManager(CAVEConfig *cfg) :
		_win            (NullFC),
	_root           (NullFC),
	_wand           (NullFC),
	_foreground     (NullFC),
	_statforeground (NullFC),
	_statstate      (false),

	_highlight      (NullFC),
	_highlightNode  (NullFC),
	_highlightPoints(NullFC),

	_internalRoot   (NullFC),
	_headlight      (NullFC),
	_litScene       (NullFC),
	_action         (NULL  ),
	_camTrolleyTrans (NullFC),

	_lastx          (TypeTraits<Int16>::getMax()),
	_lasty          (TypeTraits<Int16>::getMax()),
	_mousebuttons   (0                             ),
	_eyeseparation  (6),
	_userTransform  (NullFC),
	_sceneBoundaryMin(-2000, -2000, -2000),
	_sceneBoundaryMax(2000, 2000, 2000),
	_yRotation (0),
	_useShadows ( false )
{
	_cfg=cfg;

	for (int wallCnt=0; wallCnt<_cfg->getNumActiveWalls(); wallCnt++) {
		//    for (int camCnt=FirstCamera; camCnt<LastCamera; camCnt++) {
		_cameras[wallCnt] = NullFC;
		//        _userTransform[camCnt] = NullFC;
	}
	//    _userTransform[0] = NullFC;
	_eyeseparation = _cfg->getEyeSeparation();

/* ADDED BY LANDI */
	_scaleFactor = 1;
}


/*! pseudo constructor. Standard interface for OpenSG object creation.
 */
// CAVESceneManager* CAVESceneManager::create(bool useStereo)
// {
//     CAVESceneManager *mgr = new CAVESceneManager;
//     if (mgr) {
//         mgr->_cfg->isStereo() = useStereo;
//     }
//     return mgr;
// }

CAVESceneManager* CAVESceneManager::create(void)
{
	return new CAVESceneManager;
}


/*! Destructor
 */

CAVESceneManager::~CAVESceneManager(void)
{
	if(_action)
		delete _action;
}


/*! get the window to be used for display
 */
WindowPtr CAVESceneManager::getWindow(void)
{
	return _win;
}

/*! get the root of the displayed tree
 */
NodePtr CAVESceneManager::getRoot(void)
{
	return _root;
}

/*! get the navigator
 */
// Navigator *CAVESceneManager::getNavigator(void)
// {
//     return &_navigator;
// }

/*! get the headlight state
 */
bool CAVESceneManager::getHeadlightState(void)
{
	return _headlight->getOn();
}

/*! get the trolley translation
 */
Vec3f CAVESceneManager::getTranslation(void)
{
//#ifndef printtranslationbits
#ifdef printtranslationbits
	// tmp bits to find out innards or camera
	Matrix cam;
	_cameraTrolleyNode->getToWorld(cam);

	Quaternion Q, Q1;
	Vec3f T, T1;
	cam.getTransform(T, Q, T1, Q1);
	printf("\ntranslation = %f, %f, %f", T[0], T[1], T[2]);
#endif
	return _camTrolleyTrans->getTranslation();
}

/*! get the trolley Y rotation translation
 */
Real32 CAVESceneManager::getYRotate(void)
{
	return _yRotation;
}

/*! get the wand node
 */
NodePtr CAVESceneManager::getWand(void)
{
	return _wand;
}

/*! get the wand rotation
 *
 * @return wand rotation - amh should return NullFC idf no wand...
 */
Quaternion CAVESceneManager::getWandRotate(void)
{
	return _wandCT->getRotation();;
}


/*! set the window to be used for display
 */
void CAVESceneManager::setWindow(WindowPtr win)
{
	_win = win;
//    if(_win->getMFPort()->size() > 0 && _win->getPort(0) != NullFC)
//        _navigator.setViewport(_win->getPort(0));
}

/*! get the highlight object
 */
NodePtr CAVESceneManager::getHighlight(void)
{
	return _highlight;
}

/*! get the eye separation value
 */
Real32 CAVESceneManager::getEyeSeparation(void)
{
	return _eyeseparation;
}

/*! get the action used to render the scene
 */
DrawActionBase *CAVESceneManager::getAction(void)
{
	return _action;
}

/*! set the root of the displayed tree
 */
void CAVESceneManager::setRoot(NodePtr root)
{
	if(_internalRoot == NullFC)
	{
		initialize();
	}

	if(_root != NullFC)
	{
		_internalRoot->subChild(_root);
		subRefCP(_root);
	}

	_root = root;

    addRefCP(_root);
	_internalRoot->addChild(_root);
}

/*! set the headlight setting
 */
void CAVESceneManager::setHeadlight(bool on)
{
	if(_headlight != NullFC)
	{
		beginEditCP(_headlight, LightBase::OnFieldMask);
		_headlight->setOn(on);
		endEditCP(_headlight, LightBase::OnFieldMask);
	}
}

/*! turn headlight on.
 */
void CAVESceneManager::turnHeadlightOn(void)
{
#ifdef OSG_WIN32_ICL
#pragma warning (disable : 383)
#endif

	if(_headlight != NullFC)
	{
		beginEditCP(_headlight, LightBase::OnFieldMask);
		_headlight->setOn(true);
		endEditCP(_headlight, LightBase::OnFieldMask);
	}
#ifdef OSG_WIN32_ICL
#pragma warning (default : 383)
#endif
}

/*! turn headlight off.
 */
void CAVESceneManager::turnHeadlightOff(void)
{
#ifdef OSG_WIN32_ICL
#pragma warning (disable : 383)
#endif

	if(_headlight != NullFC)
	{
		beginEditCP(_headlight, LightBase::OnFieldMask);
		_headlight->setOn(false);
		endEditCP(_headlight, LightBase::OnFieldMask);
	}
#ifdef OSG_WIN32_ICL
#pragma warning (default : 383)
#endif
}

/*! set the trolley translation
 */
void CAVESceneManager::setTranslation(Vec3f Trans)
{
	// check that the new translation is actually allowed
	Trans = validateNewTranslation(Trans);
	// amh really I sould check return of validate here, to make sure it's not an unofficial teleport
	// although, beware cos sometimes thuis func is used for initial teleports - e.g. at startup etc.
	// for now assume anyfunc calling this has already checked validate is ok, + won't teleport (e.g.
	// in setRelativeTranslation() below...


	beginEditCP(_camTrolleyTrans, ComponentTransform::TranslationFieldMask);
		_camTrolleyTrans->setTranslation(Trans);
	endEditCP(_camTrolleyTrans, ComponentTransform::TranslationFieldMask);

}

/*! set the trolley translation, based relative to the current position and
 *  direction being faced
 */
void CAVESceneManager::setRelativeTranslation(Vec3f Trans)
{
	// get current position
	Vec3f oldTrans = getTranslation();
	// rotate new translation based on current direction being faced
	Matrix m;
	Vec3f rotatedTrans;

	m.setIdentity();
	m.setTransform(Quaternion(Vec3f(0,1,0), getYRotate()));
	m.mult(Trans, rotatedTrans);

	// calculate new transform and apply it
	rotatedTrans += oldTrans;

	setTranslation(rotatedTrans);

}

/*! set the trolley rotation
 */
void CAVESceneManager::setRotation(Quaternion Rotation)
{
	beginEditCP(_camTrolleyTrans, ComponentTransform::RotationFieldMask);
		_camTrolleyTrans->setRotation(Rotation);
	endEditCP(_camTrolleyTrans, ComponentTransform::RotationFieldMask);
}

/* ADDED BY LANDI */
void CAVESceneManager::setCameraMatrix(Matrix m)
{
	beginEditCP(_camTrolleyTrans, ComponentTransform::MatrixFieldMask);
		((TransformPtr)_camTrolleyTrans)->setMatrix(m);
	endEditCP(_camTrolleyTrans, ComponentTransform::MatrixFieldMask);
} // setCameraMatrix

/*! set a background to an image
 */
void CAVESceneManager::setBackground(Int16 position, ImagePtr image)
{
	if ((position < 0) || (position >= _cfg->getNumActiveWalls())) {
		return;
	}
	//amh this isn't the nicest ways of doing things, but works. It seems like
	// changing the background type (Image to solid) on the fly doesn't work... this
	// leaves 2 alternatives i) use a dummy (small/invis) image when want solid
	// (approach taken here), or ii) create a new set of viewports with the new
	// backgrounds (untried yet)

	ImageBackgroundPtr bkg;
	// get the image background for the relevant viewports, and set the image
	if (!_cfg->isStereo()) {
		ViewportPtr vp = _win->getPort(position);
		bkg = ImageBackgroundPtr::dcast(vp->getBackground());
		beginEditCP(bkg);
			bkg->setImage(image);
			bkg->setScale(true);
		endEditCP(bkg);
	} else {
		StereoBufferViewportPtr vp;
		for (int LRcnt=0; LRcnt<2; LRcnt++) {
			vp = StereoBufferViewportPtr::dcast(_win->getPort((position*2)+LRcnt));
			bkg = ImageBackgroundPtr::dcast(vp->getBackground());
			beginEditCP(bkg);
				bkg->setImage(image);
				bkg->setScale(true);
			endEditCP(bkg);
		}
	}
}

/*! clear all backgrounds
 */
void CAVESceneManager::clearBackgrounds(void)
{
	// create an null image
	ImagePtr NullImage = Image::create();
	beginEditCP(NullImage);
		UChar8 InvisPixel[] = {0,0,0,0};
		NullImage->set(Image::OSG_RGBA_PF,1,1,1,1,1,0,InvisPixel);
	endEditCP(NullImage);
	ImageBackgroundPtr bkg;
	// get the image background for the relevant viewports, and set the image
	for (int wallCnt=0; wallCnt<_cfg->getNumActiveWalls(); wallCnt++) {
		if (!_cfg->isStereo()) {
			ViewportPtr vp = _win->getPort(wallCnt);
			bkg = ImageBackgroundPtr::dcast(vp->getBackground());
			beginEditCP(bkg);
				bkg->setImage(NullImage);
				bkg->setScale(false);
			endEditCP(bkg);
		} else {
			StereoBufferViewportPtr vp;
			for (int LRcnt=0; LRcnt<2; LRcnt++) {
				vp = StereoBufferViewportPtr::dcast(_win->getPort((wallCnt*2)+LRcnt));
				bkg = ImageBackgroundPtr::dcast(vp->getBackground());
				beginEditCP(bkg);
					bkg->setImage(NullImage);
					bkg->setScale(false);
				endEditCP(bkg);
			}
		}
	}
}

/*! set the trolley Y rotation
 */
void CAVESceneManager::setYRotate(Real32 rotation)
{
	_yRotation = rotation;
	beginEditCP(_camTrolleyTrans, ComponentTransform::RotationFieldMask);
		_camTrolleyTrans->setRotation(Quaternion (Vec3f(0,1,0), _yRotation));
	endEditCP(_camTrolleyTrans, ComponentTransform::RotationFieldMask);
}


/*! add to the trolley Y rotation
 */
void CAVESceneManager::addYRotate(Real32 rotation)
{
	setYRotate(getYRotate()+rotation);
}

/*! set the wand geometry
 *
 * @param wandNode the nodetree of the wand to add (it's geometry, + other parameters)
 * @param SetPattern a preset pattern to use (e.g. cylinder, beam?) - default (0) = still to be decided)
 * @return -1 if already have a node present, else 0 (currently -2 for non supported preset shapes))
 */
int CAVESceneManager::setWand( NodePtr wandNode ,int setPattern)
{
	// if already have a wand, error (later might want to consider multiple wands)
 //   if (_wandCT.node() != NullFC) {
//		return -1;
//	}

	Vec3f offset(0,0,0);
	Quaternion rotation(Vec3f(0,1,0),Pi/2);

	// if no node passed in, assume a set pattern is required
	if (wandNode == NullFC) {
		switch (setPattern) {
			case 0: {

				wandNode = makeCylinder(20, 2, 8, true, true, true);
				printf("\n preset shape = like old one (cylinder)");

				// and colour it
				GeometryPtr geo = GeometryPtr::dcast(wandNode->getCore());
				SimpleMaterialPtr sm = SimpleMaterial::create();
				beginEditCP(sm, SimpleMaterial::DiffuseFieldMask);
					sm->setDiffuse(Color3f( 1, 1, 0));
				endEditCP  (sm, SimpleMaterial::DiffuseFieldMask);

				beginEditCP(geo, Geometry::MaterialFieldMask);
					geo->setMaterial(sm);
				endEditCP  (geo, Geometry::MaterialFieldMask);
			}
				break;

			case 1:
				 wandNode = makeSphere(3,6);
		  printf("\n preset shape = sphere");
				break;

			 case 2:
{
				  NodePtr wand = makeBox(200, 5,5,1,1,1);

 ComponentTransformPtr wandRotate;
 wandNode = makeCoredNode<ComponentTransform>(&wandRotate);

//	wandNode = CTNodePtr::create();

// TransformPtr tr;
//NodePtr gr = makeCoredNode<Transform>(&tr);
  // add the wand geometry to its transform node, + set it to the origin
   beginEditCP(wandRotate, ComponentTransform::RotationFieldMask);
	   wandRotate->setRotation(Quaternion(Vec3f(0,1,0), Pi/2));
   endEditCP(wandRotate, ComponentTransform::RotationFieldMask);

   beginEditCP(wandNode, Node::ChildrenFieldMask);
	   wandNode->addChild(wand);
   endEditCP(wandNode, Node::ChildrenFieldMask);

}

		  printf("\n preset shape = ray");
				break;

		   default:
				printf("\n preset shape %d not yet supported", setPattern);
				return -2;
		}
	}

 //for old backward compatability...
 _wand = wandNode;

   // create the transformation cored node for the wand
	_wandCT = CTNodePtr::create();
//	_wandCT = CoredNodePtr<ComponentTransform>::create();

   beginEditCP(_wandCT, ComponentTransform::CenterFieldMask );
	   _wandCT->setCenter(-offset);
   endEditCP(_wandCT, ComponentTransform::CenterFieldMask );


	// add the wand geometry to its transform node, + set it to the origin
   beginEditCP(_wandCT, ComponentTransform::TranslationFieldMask |
			ComponentTransform::RotationFieldMask |
			Node::ChildrenFieldMask);
//	   _wandCT->setTranslation(Vec3f(0,0,0));
//	   _wandCT->setRotation(Quaternion(Vec3f(0,0,0), 0));
	   _wandCT->setTranslation(offset);
	   _wandCT->setRotation(rotation);
	   _wandCT.node()->addChild(wandNode);
   endEditCP(_wandCT, ComponentTransform::TranslationFieldMask |
			ComponentTransform::RotationFieldMask |
			Node::ChildrenFieldMask);

	// and add it to the trolley
   beginEditCP(_cameraTrolleyNode, Node::ChildrenFieldMask );
	   _cameraTrolleyNode->addChild(_wandCT.node());
   endEditCP(_cameraTrolleyNode, Node::ChildrenFieldMask );

	return 0;
}


/*! set the wand translation
 */
void CAVESceneManager::setWandTranslation(Vec3f Trans)
{
	if (_wandCT.node() == NullFC){
		return;
	}
//	beginEditCP(_wandTransform, ComponentTransform::TranslationFieldMask);
//		_wandTransform->setTranslation(Trans);
//	endEditCP(_wandTransform, ComponentTransform::TranslationFieldMask);

	beginEditCP(_wandCT, ComponentTransform::TranslationFieldMask);
		_wandCT->setTranslation(Trans);
	endEditCP(_wandCT, ComponentTransform::TranslationFieldMask);
	//also change the centre, since we're rotating around the user
//   beginEditCP(_wandCT, ComponentTransform::CenterFieldMask );
//	   _wandCT->setCenter(-Trans);
//   endEditCP(_wandCT, ComponentTransform::CenterFieldMask );
}

/*! set the wand color
 */
void CAVESceneManager::setWandColour(Color3f colour)
{
	if (_wand == NullFC){
		return;
	}

	GeometryPtr geo = GeometryPtr::dcast(_wand->getCore());

	SimpleMaterialPtr sm = SimpleMaterialPtr::dcast(geo->getMaterial());
	beginEditCP(sm, SimpleMaterial::DiffuseFieldMask);
		sm->setDiffuse(colour);
	endEditCP  (sm, SimpleMaterial::DiffuseFieldMask);
}

/*! set the wand rotation
 *
 * @param rotate angle to set wand rotation to
 * @return void
 */
 void CAVESceneManager::setWandRotate (Quaternion rotation )
{
   beginEditCP(_wandCT, ComponentTransform::RotationFieldMask);
	   _wandCT->setRotation(rotation);
   endEditCP(_wandCT, ComponentTransform::RotationFieldMask);
}

/*! rotate the wand
 *
 * @param rotate angle to rotate by
 * @return void
 */
 void CAVESceneManager::rotateWand (Quaternion rotation )
{
	Quaternion q = _wandCT->getRotation();
	q.mult(rotation);
	setWandRotate(q);
}


/*! turn wand on.
 */
void CAVESceneManager::turnWandOn(void)
{
	CreateWand(_cameraTrolleyNode);
}

/*! turn wand off.
 */
void CAVESceneManager::turnWandOff(void)
{
	if(_wand != NullFC)
	{
		// is this next sufficient for clearing up? smart pointers should take care of mem???
		beginEditCP(_cameraTrolleyNode, Node::ChildrenFieldMask );
			_cameraTrolleyNode->subChild(_wandTransformNode);
		endEditCP(_cameraTrolleyNode, Node::ChildrenFieldMask );
/*		beginEditCP(_wandTransformNode, Node::ChildrenFieldMask );
			_wandTransformNode->subChild(_wand);
		endEditCP(_wandTransformNode, Node::ChildrenFieldMask );*/
		_wandTransformNode = NullFC;
		_wand = NullFC;
	}
}

/*! set the user transform
 */
void  CAVESceneManager::setUserTransform ( Vec3f trans, Quaternion rotate )
{
	Matrix m;

/* ADDED BY LANDI */
//	 trans *= _scaleFactor;

	beginEditCP(_userTransform, Transform::MatrixFieldMask);
		m.setIdentity();
		m.setTransform(trans, rotate);
		_userTransform->setMatrix(m);
	endEditCP (_userTransform, Transform::MatrixFieldMask);

	// check to see if the user transform takes the user out of bounds (in which
	// case pull them back through the translation)
	Vec3f currentTrans = _camTrolleyTrans->getTranslation();
/*
	Vec3f postValidatedTrans  = validateNewTranslation(currentTrans);



	if (postValidatedTrans != currentTrans) {
//for now ignore setting trans - it seems to be causing the jumping
printf("\n\nwould have teleported to %f, %f, %f - instead of %f, %f, %f\n",
	postValidatedTrans[0], postValidatedTrans[1],postValidatedTrans[2],
	currentTrans[0], currentTrans[1], currentTrans[2]);
//		setTranslation(postValidatedTrans);
	}*/

}

void  CAVESceneManager::setSceneBoundaries ( Vec3f Min, Vec3f Max )
{
	_sceneBoundaryMin = Min;
	_sceneBoundaryMax = Max;
}


/*! set the highlight object
 */
void CAVESceneManager::setHighlight(NodePtr highlight)
{
	_highlight = highlight;
	highlightChanged();
}

/*! set the eye separation value
 */
void CAVESceneManager::setEyeSeparation(Real32 eyeSep)
{
	if ((!_cfg->isStereo()) || (_win == NullFC)) {
		return;
	}
	_eyeseparation = eyeSep;
	ProjectionCameraDecoratorPtr CamDec;
	for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++) {
		CamDec = ProjectionCameraDecoratorPtr::dcast(_win->getPort(portCnt)->getCamera());
		beginEditCP(CamDec, ProjectionCameraDecorator::EyeSeparationFieldMask );
			CamDec->setEyeSeparation(_eyeseparation * _scaleFactor);
		endEditCP  (CamDec, ProjectionCameraDecorator::EyeSeparationFieldMask );
	}
}


/*! set the statistics setting. Only works after the window has been created.
 */
void CAVESceneManager::setStatistics(bool on)
{
	if(_statforeground != NullFC && on != _statstate)
	{
		if(on)
		{
			printf("\nCSM putting stats on");
			if (!_cfg->isStereo()) {
				ViewportPtr vp = (_win->getPort(0));
				vp->getForegrounds().push_back(_statforeground);

			} else {
				StereoBufferViewportPtr vp = StereoBufferViewportPtr::dcast(_win->getPort(0));
				vp->getForegrounds().push_back(_statforeground);
				vp = StereoBufferViewportPtr::dcast(_win->getPort(1));
				vp->getForegrounds().push_back(_statforeground);

			}

			_action->setStatistics(&_statforeground->getCollector());
		}
		else
		{
				printf("\nCSM putting stats off");
			MFForegroundPtr::iterator it;
			if (!_cfg->isStereo()) {
				ViewportPtr vp = (_win->getPort(0));
				it = vp->getForegrounds().find(_statforeground);
				vp->getForegrounds().erase(it);

			} else {
				StereoBufferViewportPtr vp = StereoBufferViewportPtr::dcast(_win->getPort(0));
				it = vp->getForegrounds().find(_statforeground);
				vp->getForegrounds().erase(it);
				vp = StereoBufferViewportPtr::dcast(_win->getPort(1));
				it = vp->getForegrounds().find(_statforeground);
				vp->getForegrounds().erase(it);

			}

			_action->setStatistics(NULL);
		}

		_statstate = on;
	}
}

void CAVESceneManager::setForeground(Int16 position, ImageForegroundPtr fg, bool clear)
{
	if ((position < 0) || (position >= _cfg->getNumActiveWalls())) {
		return;
	}

	if(fg != NullFC)
	{

//		StereoBufferViewportPtr vp = _win->getPort()[0];
//		ViewportPtr vp = _win->getPort()[0];
//		StereoBufferViewportPtr vp = StereoBufferViewportPtr::dcast(_win->getPort(0));
		if (!_cfg->isStereo()) {
			// 2007-07-18 ZaJ: use position parameter
			//ViewportPtr vp = (_win->getPort(0));
			ViewportPtr vp = (_win->getPort(position));
			beginEditCP(vp);
				if (clear) vp->getForegrounds().clear();
				vp->getForegrounds().push_back(fg);
			endEditCP(vp);

			//vp->render(_action);
		} else {
			StereoBufferViewportPtr vp;
			for (int LRcnt=0; LRcnt<2; LRcnt++) {
				// 2007-07-18 ZaJ: use position-information:
				vp = StereoBufferViewportPtr::dcast(_win->getPort((position*2)+LRcnt));
				beginEditCP(vp);
					if (clear) vp->getForegrounds().clear();
					vp->getForegrounds().push_back(fg);
				endEditCP(vp);
			}

			//vp->render(_action);
		}

			//_action->setStatistics(&fg->getCollector());

	}
}

/*-------------------------------------------------------------------------*/
/*							   Updates								   */

/*! create the parts that the app needs: viewport, camera, base graph etc.
 */
void CAVESceneManager::initialize(void)
{
	// the rendering action
	_action = RenderAction::create();

	// the lit scene (later if needed, could
	// change its core to something other than group (comp transform?))
	_litScene = Node::create();
	GroupPtr litGroup = Group::create();
	beginEditCP(_litScene, Node::CoreFieldMask);
		_litScene->setCore(litGroup);
	endEditCP (_litScene, Node::CoreFieldMask);

	// the camera trolley
	_cameraTrolleyNode = Node::create();
	_camTrolleyTrans = ComponentTransform::create();
	beginEditCP(_camTrolleyTrans, ComponentTransform::TranslationFieldMask |
				ComponentTransform::RotationFieldMask);
		_camTrolleyTrans->setTranslation(Vec3f(0,0,0));
		_camTrolleyTrans->setRotation(Quaternion(Vec3f(0,0,0), 0));
	endEditCP(_camTrolleyTrans, ComponentTransform::TranslationFieldMask |
				ComponentTransform::RotationFieldMask);

	// create the user translations (where as user's position is in the cave)
	_userTransform = Transform::create();
	beginEditCP(_userTransform, Transform::MatrixFieldMask);
		_userTransform->setMatrix(Matrix::identity());
	endEditCP (_userTransform, Transform::MatrixFieldMask);

	NodePtr user = Node::create();
	beginEditCP(user, Node::CoreFieldMask);
		user->setCore(_userTransform);
	endEditCP (user, Node::CoreFieldMask);

	// put the camera becons on the trolley
	beginEditCP(_cameraTrolleyNode, Node::ChildrenFieldMask | Node::CoreFieldMask );
		_cameraTrolleyNode->setCore(_camTrolleyTrans);
		_cameraTrolleyNode->addChild(user);
	endEditCP(_cameraTrolleyNode, Node::ChildrenFieldMask | Node::CoreFieldMask );


	// the headlight, which is also the internal root (and parent of the lit scene)
	_internalRoot = Node::create();
	_headlight	= DirectionalLight::create();

	addRefCP(_internalRoot);
	beginEditCP(_internalRoot, Node::ChildrenFieldMask | Node::CoreFieldMask );
		_internalRoot->setCore(_headlight);
		_internalRoot->addChild(_litScene);
	endEditCP(_internalRoot, Node::ChildrenFieldMask | Node::CoreFieldMask );

	// set up headlight, and point it to the lit scene
	beginEditCP(_headlight);
//		_headlight->setAmbient  (.7, .7, .7, 1);
//		_headlight->setDiffuse  (.7, .7, .7, 1);
//		_headlight->setSpecular (.9, .9, .9, 1);
//		_headlight->setAmbient  (0.2,0.2,0.2, 1);
//		_headlight->setDiffuse  (1,1,1, 1);
//		_headlight->setSpecular (1,1,1, 1);
		_headlight->setAmbient  (0.9,0.8,0.8, 1);
		_headlight->setDiffuse  (0.6,0.6,0.6, 1);
		_headlight->setSpecular (1,1,1, 1);
		_headlight->setDirection( 0,  0,  1);
//		_headlight->setDirection( 0,  0,  1);
//		_headlight->setBeacon   (_litScene);
		_headlight->setBeacon   (user);
	endEditCP(_headlight);


	// put the cameras on the lit scene
	beginEditCP(_litScene, Node::ChildrenFieldMask);
		_litScene->addChild(_cameraTrolleyNode);
	endEditCP (_litScene, Node::ChildrenFieldMask);

	// create the cameras
	for (int wallCnt=0; wallCnt<_cfg->getNumActiveWalls(); wallCnt++) {
		_cameras[wallCnt] = PerspectiveCamera::create();
		beginEditCP(_cameras[wallCnt]);
			_cameras[wallCnt]->setBeacon(_cameraTrolleyNode);
			_cameras[wallCnt]->setFov   (deg2rad(90.f));
			_cameras[wallCnt]->setNear  (0.003f);
			_cameras[wallCnt]->setFar   (1000.f);
		endEditCP(_cameras[wallCnt]);
	}

	CreateWand(_cameraTrolleyNode);

	// need a viewport?
	if(_win != NullFC && _win->getPort().size() == 0)
	{

		_foreground = ImageForeground::create();

		SimpleStatisticsForegroundPtr sf = SimpleStatisticsForeground::create();

		beginEditCP(sf);
			sf->setSize(25);
			sf->setColor(Color4f(0,1,0,0.7));

			sf->addElement(DrawActionBase::statCullTestedNodes,
							"%d Nodes culltested");


			sf->addElement(RenderAction::statTravTime,	  "Draw FPS: %r.3f");


			sf->addElement(DrawActionBase::statCulledNodes,
							"%d Nodes culled");
			sf->addElement(RenderAction::statNMaterials,
							"%d material changes");
			sf->addElement(RenderAction::statNMatrices,
							"%d matrix changes");
			sf->addElement(RenderAction::statNGeometries,
							"%d Nodes drawn");
			sf->addElement(RenderAction::statNTransGeometries,
							"%d transparent Nodes drawn");
#if 0 // not ready for primetime yet
			sf->addElement(PointLight::statNPointLights,
							"%d active point lights");
			sf->addElement(DirectionalLight::statNDirectionalLights,
							"%d active directional lights");
			sf->addElement(SpotLight::statNSpotLights,
							"%d active spot lights");
#endif
			sf->addElement(Drawable::statNTriangles,	"%d triangles drawn");
			sf->addElement(Drawable::statNLines,		"%d lines drawn");
			sf->addElement(Drawable::statNPoints,	   "%d points drawn");
			sf->addElement(Drawable::statNVertices,	 "%d vertices transformed");
		endEditCP(sf);

		StatCollector *collector = &sf->getCollector();

		// add optional elements
		collector->getElem(Drawable::statNTriangles);
//		collector->getElem(amhStatDesc);

		_statforeground = sf;


		// construct array of where each walls image will be rendered
		Pnt4f vppos[MAXCAVEWALLS];
		// work out number of pipes being used (highest pipe index+1)
		float PipeDevisionSize = 1.0/_cfg->getNumPipes();
		for (int wallCnt=0; wallCnt<_cfg->getNumActiveWalls(); wallCnt++) {
			CAVEWall* wall = _cfg->getActiveWall(wallCnt);
			// if no geometry for this wall, treat it as full size, and continue to next wall
			if (!wall->getGeometry()) {
				float startPos = wall->getPipeIndex() * PipeDevisionSize;
				vppos[wallCnt] = Pnt4f( startPos, 0, startPos+PipeDevisionSize, 1);
				continue;
			}


			float MaxX=0, MaxY=0;
			// work out MAX resolution for other pipes with this index
			for (int i=0; i<_cfg->getNumActiveWalls(); i++) {
				if (wall->getPipeIndex() == _cfg->getActiveWall(i)->getPipeIndex()) {
					// if there is no geometry for this pipe, don't count it - cont to next
					if (!_cfg->getActiveWall(i)->getGeometry()) {
						continue;
					}
					// see if this ones geom is bigger than current X,Y - use it if so.
					if (MaxX < _cfg->getActiveWall(i)->getGeomCoordMaxX()) {
						MaxX = _cfg->getActiveWall(i)->getGeomCoordMaxX();
					}
					if (MaxY < _cfg->getActiveWall(i)->getGeomCoordMaxY()) {
						MaxY = _cfg->getActiveWall(i)->getGeomCoordMaxY();
					}
				}
			}
			// work out positions for this wall within the pipe
			vppos[wallCnt] = Pnt4f( (wall->getPipeIndex() + ((float)wall->getGeomCoordMinX()/MaxX)) * PipeDevisionSize,
					(float)wall->getGeomCoordMinY()/MaxY,
					(wall->getPipeIndex() + ((float)wall->getGeomCoordMaxX()/MaxX)) * PipeDevisionSize,
					(float)wall->getGeomCoordMaxY()/MaxY);
		}

		printf("\nWallParams");
		for (int wallCnt=0; wallCnt<_cfg->getNumActiveWalls(); wallCnt++) {
			printf("\nWall %d, pipeIdx %d  %.4f, %.4f, %.4f, %.4f", wallCnt,
					_cfg->getActiveWall(wallCnt)->getPipeIndex(),
					vppos[wallCnt][0], vppos[wallCnt][1], vppos[wallCnt][2], vppos[wallCnt][3]);
		}



		// create a NULL image for the backgrounds
		ImagePtr NullImage = Image::create();
		beginEditCP(NullImage);
			UChar8 InvisPixel[] = {0,0,0,0};
			NullImage->set(Image::OSG_RGBA_PF,1,1,1,1,1,0,InvisPixel);
		endEditCP(NullImage);
		ImageBackgroundPtr bkg;
		 // create 4/8 viewports - number + type depending on if we're using stereo or not...


		for (int wallCnt=0; wallCnt<_cfg->getNumActiveWalls(); wallCnt++) {
			CAVEWall* wall;
			if (!(wall = _cfg->getActiveWall(wallCnt))) {
				printf("\n\nSkipping Wall %d\n\n",wallCnt);
				continue;
			}
			bkg = ImageBackground::create();
			beginEditCP(bkg);
				bkg->setColor(Color3f(0, 0, 0));
				bkg->setImage(NullImage);
				bkg->setScale(false);
			endEditCP(bkg);
			if (!_cfg->isStereo()) {
				if (!_useShadows)
				{
					ViewportPtr vp = CreateMonoProjCamVeiwport(_cameras[wallCnt],
						vppos[wallCnt], bkg, user, wall->getProjData());
					beginEditCP(_win);
						_win->addPort(vp);
					endEditCP(_win);
				} else {
					ShadowViewportPtr vp = CreateMonoShadowViewport(_cameras[wallCnt],
						vppos[wallCnt], bkg, user, wall->getProjData());
					beginEditCP(_win);
						_win->addPort(vp);
					endEditCP(_win);
				}
			} else {
//printf("\n\nWall %d",wallCnt);
//Pnt3f pnt;
//printf("\n\twall data =");
//for (int i=0; i<4; i++) {
//	pnt=*(wall->getProjData()+i);
//	printf("%f,%f,%f  ",pnt[0], pnt[1], pnt[2]);
//}
				// Shadow Support
				if(!_useShadows)
				{
					StereoBufferViewportPtr vp;
					for (int LRcnt=0; LRcnt<2; LRcnt++)
					{
						vp = CreateProjStereoVeiwport(LRcnt?false:true, _cameras[wallCnt],
								vppos[wallCnt], bkg, user, wall->getProjData());
//								vppos[wallCnt], bkg, user, vpproj[wallCnt]);
						beginEditCP(_win);
							_win->addPort(vp);
						endEditCP(_win);
					}
		 		}
		 		else
		 		{
					ShadowViewportPtr vp;
					for (int LRcnt=0; LRcnt<2; LRcnt++)
					{
						vp = CreateStereoShadowViewport(LRcnt?false:true, _cameras[wallCnt],
								vppos[wallCnt], bkg, user, wall->getProjData());
//								vppos[wallCnt], bkg, user, vpproj[wallCnt]);
						beginEditCP(_win);
							_win->addPort(vp);
						endEditCP(_win);
					}
				}

			}
		}
	}

}

/*-------------------------------------------------------------------------*/
/*				Manipulating other objects in the scene				  */

/*! creating external scene objects etc
 */

bool  CAVESceneManager::addToScene(NodePtr node, Int16 flags )
{
	// could check that passed in node is valid - i.e. doesn't already have parent etc.
	// don't use flags yet, but just refernce them to help the build!
	if (flags) {
		printf("\nCSM flags %d",flags);
	}

	// create a transform node for the object to add
	TransformPtr nodeTransform = Transform::create();
	beginEditCP(nodeTransform, Transform::MatrixFieldMask);
		Matrix m;
		m.setIdentity();
		nodeTransform->setMatrix(m);
	endEditCP (nodeTransform, Transform::MatrixFieldMask);

	// create a node for the transform. Set its core and child nodes
	NodePtr addedTransNode = Node::create();
	beginEditCP(addedTransNode, Node::ChildrenFieldMask | Node::CoreFieldMask );
		addedTransNode->setCore(nodeTransform);
		addedTransNode->addChild(node);
	endEditCP(addedTransNode, Node::ChildrenFieldMask | Node::CoreFieldMask );
	// and add the tranceform node to the lit scene
	beginEditCP(_litScene, Node::ChildrenFieldMask);
		_litScene->addChild(addedTransNode);
//		_litScene->addChild(node);
	endEditCP (_litScene, Node::ChildrenFieldMask);
	return true;
}


bool CAVESceneManager::removeFromScene(NodePtr node )
{
	// could check that passed in node is valid - i.e. that it already belongs
	// to the litscene etc

	// find the parent of passed in node
	NodePtr parent = node->getParent();
	if ((parent==NullFC) ||
			(parent->getCore()==NullFC) ||
			(parent->getCore()->getTypeId() != Transform::getClassTypeId())) {
		// invalid node passed in
		return false;
	}
	// delete the transform
	// is this next sufficient for clearing up? smart pointers should take care of
	// also removing the node itself, thereby freeing the extra mem???
	beginEditCP(_litScene, Node::ChildrenFieldMask );
		_litScene->subChild(parent);
	endEditCP(_litScene, Node::ChildrenFieldMask );
	return true;
}


bool CAVESceneManager::setTranslationInScene(NodePtr node, Vec3f trans, Quaternion rotate )
{
	// could check that passed in node is valid - i.e. that it already belongs
	// to the litscene etc

	// find the parent of passed in node
	NodePtr parent = node->getParent();
	if ((parent==NullFC) ||
			(parent->getCore()==NullFC) ||
			(parent->getCore()->getTypeId() != Transform::getClassTypeId())) {
		// invalid node passed in
		return false;
	}
	TransformPtr parentTrans = TransformPtr::dcast(parent->getCore());
	// translate by required amount
	Matrix m;
	beginEditCP(parentTrans, Transform::MatrixFieldMask);
		m.setIdentity();
		m.setTransform(trans, rotate);
		parentTrans->setMatrix(m);
	endEditCP (parentTrans, Transform::MatrixFieldMask);
	return true;
}

/*-------------------------------------------------------------------------*/
/*				Manipulating other objects in the scene				  */

/*! attaching/manipulating external scene objects wrt the camera trolley
 */

/*! attach an object to the trolley
 * @param node node to be attached to trolley
 * @param flags flags relating to how to add
 * @return true is sucessfully added
 */
bool  CAVESceneManager::addToTrolley(NodePtr node, Int16 flags )
{
	// could check that passed in node is valid - i.e. doesn't already have parent etc.
	// don't use flags yet, but just refernce them to help the build!
	if (flags) {
		printf("\nCSM flags %d",flags);
	}

	// create a transform node for the object to add
// amh - do this with cored nodes
	TransformPtr nodeTransform = Transform::create();
	beginEditCP(nodeTransform, Transform::MatrixFieldMask);
		Matrix m;
		m.setIdentity();
		nodeTransform->setMatrix(m);
	endEditCP (nodeTransform, Transform::MatrixFieldMask);

	// create a node for the transform. Set its core and child nodes
	NodePtr addedTransNode = Node::create();
	beginEditCP(addedTransNode, Node::ChildrenFieldMask | Node::CoreFieldMask );
		addedTransNode->setCore(nodeTransform);
		addedTransNode->addChild(node);
	endEditCP(addedTransNode, Node::ChildrenFieldMask | Node::CoreFieldMask );
	// and add the tranceform node to the camera trolley node
	beginEditCP(_cameraTrolleyNode, Node::ChildrenFieldMask);
		_cameraTrolleyNode->addChild(addedTransNode);
//		_litScene->addChild(node);
	endEditCP (_cameraTrolleyNode, Node::ChildrenFieldMask);
	return true;
}


/*! remove an attached object from the trolley
 * @param node node to be removed
 * @return true is sucessfully removed, false if couldn't find node
 */
bool CAVESceneManager::removeFromTrolley(NodePtr node )
{
	// could check that passed in node is valid - i.e. that it already belongs
	// to the litscene etc

	// find the parent of passed in node
	NodePtr parent = node->getParent();
	if ((parent==NullFC) ||
			(parent->getCore()==NullFC) ||
			(parent->getCore()->getTypeId() != Transform::getClassTypeId())) {
		// invalid node passed in
		return false;
	}
	// delete the transform
	// is this next sufficient for clearing up? smart pointers should take care of
	// also removing the node itself, thereby freeing the extra mem???
	beginEditCP(_cameraTrolleyNode, Node::ChildrenFieldMask );
		_cameraTrolleyNode->subChild(parent);
	endEditCP(_cameraTrolleyNode, Node::ChildrenFieldMask );
	return true;
}


/*! set the translation of a node attached to the trolley -
 * @param node node to be moved
 * @param trans translation of object wrt trolley
 * @param rotate rotation of object wrt trolley
 * @return true is sucessfully moved, false if couldn't find node
 */
bool CAVESceneManager::setTranslationFromTrolley(NodePtr node, Vec3f trans, Quaternion rotate )
{
	// could check that passed in node is valid - i.e. that it already belongs
	// to the litscene etc

	// find the parent of passed in node
	NodePtr parent = node->getParent();
	if ((parent==NullFC) ||
			(parent->getCore()==NullFC) ||
			(parent->getCore()->getTypeId() != Transform::getClassTypeId())) {
		// invalid node passed in
		return false;
	}
	TransformPtr parentTrans = TransformPtr::dcast(parent->getCore());
	// translate by required amount
	Matrix m;
	beginEditCP(parentTrans, Transform::MatrixFieldMask);
		m.setIdentity();
		m.setTransform(trans, rotate);
		parentTrans->setMatrix(m);
	endEditCP (parentTrans, Transform::MatrixFieldMask);
	return true;
}


/*! validate if a new trnaslation position is valid. If not the
 *  last valid point on a line drawn from the current position to
 * the desired translation is used
 */
Vec3f CAVESceneManager::validateNewTranslation(Vec3f desiredTrans)
{
	// add local user position (in CAVE) to current and desired position
	Matrix m = _userTransform->getMatrix();
	Vec3f localTrans = Vec3f(m[3][0], m[3][1], m[3][2]);

if ((fabsf(localTrans[0])>1000) ||(fabsf(localTrans[1])>1000) ||(fabsf(localTrans[2])>1000)) {
	printf("\n\nWeird localTrans %f, %f, %f\n",localTrans[0], localTrans[1], localTrans[2]);
}
//	Vec3f currentTranslation = _camTrolleyTrans->getTranslation() + localTrans;
	Vec3f desiredLocation = desiredTrans + localTrans;
	// later could add collision detection between current + desired
	// to find if any blocking planes are crossed...  for now just use sizes
	for (int i=0; i<3; i++) {
		if (desiredLocation[i] > _sceneBoundaryMax[i]) {
			desiredLocation[i] = _sceneBoundaryMax[i];
		} else if (desiredLocation[i] < _sceneBoundaryMin[i]) {
			desiredLocation[i] = _sceneBoundaryMin[i];
		}
	}
	desiredLocation -= localTrans;

	return desiredLocation;

}


/*! show the whole scene: move out far enough  to see everything
 */
void CAVESceneManager::showAll(void)
{
	if(_root == NullFC)
		return;

	_root->updateVolume();

	Vec3f min,max;
	_root->getVolume().getBounds( min, max );
	Vec3f d = max - min;

	Real32 dist = osgMax(d[0],d[1]) / (2 * osgtan(_cameras[0]->getFov() / 2.f));

	Vec3f up(0,1,0);
	Pnt3f at((min[0] + max[0]) * .5,(min[1] + max[1]) * .5,(min[2] + max[2]) * .5);
	Pnt3f from=at;
	from[2]-=dist;

	// set the camera to go from 1% of the object to twice(??) its size

	Real32 diag = osgMax(osgMax(d[0], d[1]), d[2]);
	for (int wallCnt=0; wallCnt<_cfg->getNumActiveWalls(); wallCnt++) {
//	for ( int camCnt=FirstCamera; camCnt<LastCamera; camCnt++) {
		beginEditCP(_cameras[wallCnt]);
			_cameras[wallCnt]->setNear (diag / 3000.f);
			_cameras[wallCnt]->setFar  (30 * diag);
		endEditCP(_cameras[wallCnt]);
	}

}

// ADDED BY HBRESS
void CAVESceneManager::setFarClippingPlane(float farDist)
{
	if(_root == NullFC)
		return;

	for (int wallCnt=0; wallCnt<_cfg->getNumActiveWalls(); wallCnt++) {
//	for ( int camCnt=FirstCamera; camCnt<LastCamera; camCnt++) {
		beginEditCP(_cameras[wallCnt]);
			_cameras[wallCnt]->setFar(farDist);
		endEditCP(_cameras[wallCnt]);
	}
}

// ADDED BY LANDI
void CAVESceneManager::setNearClippingPlane(float nearDist)
{
	if(_root == NullFC)
		return;

	for (int wallCnt=0; wallCnt<_cfg->getNumActiveWalls(); wallCnt++) {
//	for ( int camCnt=FirstCamera; camCnt<LastCamera; camCnt++) {
		beginEditCP(_cameras[wallCnt]);
			_cameras[wallCnt]->setNear(nearDist);
		endEditCP(_cameras[wallCnt]);
	}
}

// ADDED BY HBRESS,LANDI
void CAVESceneManager::setScale(float scale)
{
	int i;
	ProjectionCameraDecoratorPtr CamDec;

	for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++)
	{
		CamDec = ProjectionCameraDecoratorPtr::dcast(_win->getPort(portCnt)->getCamera());
		beginEditCP(CamDec, ProjectionCameraDecorator::SurfaceFieldMask | ProjectionCameraDecorator::EyeSeparationFieldMask);
			for (i=0; i < 4; i++)
				CamDec->getSurface(i) *= scale/_scaleFactor;

			CamDec->setEyeSeparation(_eyeseparation * scale);
		endEditCP(CamDec, ProjectionCameraDecorator::SurfaceFieldMask | ProjectionCameraDecorator::EyeSeparationFieldMask);
	} // for

	_scaleFactor = scale;

} // setScale

/*! add the "Powered by OpenSG" logo to the lower left corner
 */
void CAVESceneManager::useOpenSGLogo(void)
{
	ImagePtr lo = Image::create();
	ImageFileType::restore( lo, (UChar8*)LogoData, -1 );

	beginEditCP(_foreground);
	_foreground->addImage( lo, Pnt2f( 0,0 ) );
	endEditCP  (_foreground);
}


/*! Draw the next frame, update if needed.
 */
void CAVESceneManager::idle(void)
{
}

/*! Draw the next frame, update if needed.
 */
void CAVESceneManager::redraw(void)
{
	if (_internalRoot == NullFC)
	{
		initialize();
		showAll();
	}

	updateHighlight();

	_win->render(_action);
}

/*! Adjust the highlight to a changed object. Also initializes the highlight
   statics if called the first time.
 */
void CAVESceneManager::highlightChanged(void)
{

  // FDEBUG (("SimpleSceneManager::updateHightlight() called\n"));

	// init as needed
	if(_highlightMaterial == NullFC)
	{
		_highlightMaterial = SimpleMaterial::create();

		beginEditCP(_highlightMaterial);
		_highlightMaterial->setDiffuse (Color3f(0,1,0));
		_highlightMaterial->setLit	 (false);
		endEditCP(_highlightMaterial);
	}
	if(_highlightNode == NullFC)
	{
		GeoPTypesPtr type = GeoPTypesUI8::create();
		beginEditCP(type);
		type->push_back(GL_LINE_STRIP);
		type->push_back(GL_LINES);
		endEditCP(type);

		GeoPLengthsPtr lens = GeoPLengthsUI32::create();
		beginEditCP(lens);
		lens->push_back(10);
		lens->push_back(6);
		endEditCP(lens);

		GeoIndicesUI32Ptr index = GeoIndicesUI32::create();
		beginEditCP(index);
		index->getFieldPtr()->push_back(0);
		index->getFieldPtr()->push_back(1);
		index->getFieldPtr()->push_back(3);
		index->getFieldPtr()->push_back(2);
		index->getFieldPtr()->push_back(0);
		index->getFieldPtr()->push_back(4);
		index->getFieldPtr()->push_back(5);
		index->getFieldPtr()->push_back(7);
		index->getFieldPtr()->push_back(6);
		index->getFieldPtr()->push_back(4);

		index->getFieldPtr()->push_back(1);
		index->getFieldPtr()->push_back(5);
		index->getFieldPtr()->push_back(2);
		index->getFieldPtr()->push_back(6);
		index->getFieldPtr()->push_back(3);
		index->getFieldPtr()->push_back(7);
		endEditCP(index);

		_highlightPoints = GeoPositions3f::create();
		beginEditCP(_highlightPoints);
		_highlightPoints->push_back(Pnt3f(-1, -1, -1));
		_highlightPoints->push_back(Pnt3f( 1, -1, -1));
		_highlightPoints->push_back(Pnt3f(-1,  1, -1));
		_highlightPoints->push_back(Pnt3f( 1,  1, -1));
		_highlightPoints->push_back(Pnt3f(-1, -1,  1));
		_highlightPoints->push_back(Pnt3f( 1, -1,  1));
		_highlightPoints->push_back(Pnt3f(-1,  1,  1));
		_highlightPoints->push_back(Pnt3f( 1,  1,  1));
		endEditCP(_highlightPoints);

		GeometryPtr geo=Geometry::create();
		beginEditCP(geo);
		geo->setTypes	 (type);
		geo->setLengths   (lens);
		geo->setIndices   (index);
		geo->setPositions (_highlightPoints);
		geo->setMaterial  (_highlightMaterial);
		endEditCP(geo);
		addRefCP(geo);

		_highlightNode = Node::create();
		beginEditCP(_highlightNode);
		_highlightNode->setCore(geo);
		endEditCP(_highlightNode);
		addRefCP(_highlightNode);
	}

	// attach the hightlight node to the root if the highlight is active
	if(getHighlight() != NullFC)
	{
		if(_highlightNode->getParent() == NullFC)
		{
			beginEditCP(_internalRoot);
			_internalRoot->addChild(_highlightNode);
			endEditCP(_internalRoot);
		}
	}
	else
	{
		if(_highlightNode->getParent() != NullFC)
		{
			beginEditCP(_internalRoot);
			_internalRoot->subChild(_highlightNode);
			endEditCP(_internalRoot);
		}

	}
	// update the highlight geometry
	updateHighlight();
}

/*! Update the highlight for a moved object. Does not handle changing the
	object, this is done by highlightChanged().
 */
void CAVESceneManager::updateHighlight(void)
{
	if(_highlight==NullFC)
		return;

	// FDEBUG (("SimpleSceneManager::updateHightlight() called\n"));

	// calc the world bbox of the highlight object
	DynamicVolume vol;
	_highlight->getWorldVolume(vol);

	Pnt3f min,max;
	vol.getBounds(min, max);

	beginEditCP(_highlightPoints);
		_highlightPoints->setValue(Pnt3f(min[0], min[1], min[2]), 0);
		_highlightPoints->setValue(Pnt3f(max[0], min[1], min[2]), 1);
		_highlightPoints->setValue(Pnt3f(min[0], max[1], min[2]), 2);
		_highlightPoints->setValue(Pnt3f(max[0], max[1], min[2]), 3);
		_highlightPoints->setValue(Pnt3f(min[0], min[1], max[2]), 4);
		_highlightPoints->setValue(Pnt3f(max[0], min[1], max[2]), 5);
		_highlightPoints->setValue(Pnt3f(min[0], max[1], max[2]), 6);
		_highlightPoints->setValue(Pnt3f(max[0], max[1], max[2]), 7);
	endEditCP(_highlightPoints);

	beginEditCP(_highlightNode->getCore(), Geometry::PositionsFieldMask);
	endEditCP  (_highlightNode->getCore(), Geometry::PositionsFieldMask);
}

/*-------------------------------------------------------------------------*/
/*						Interaction handling							 */

/*! resize
 */
void CAVESceneManager::resize(UInt16 width, UInt16 height)
{
	_win->resize(width, height);
}

/*! motion
 */
void CAVESceneManager::mouseMove(Int16 x, Int16 y)
{
//	if ( _mousebuttons) _navigator.moveTo(x,y);
	_lastx = x;
	_lasty = y;

}

/*! call when a mouse button is pressed. button is the number of the pressed
	button, starting at 0, ordered from left to right. A wheel should be
	mapped to buttons 3 and 4. The position is in pixel, starting at the
	upper left corner.
 */

void CAVESceneManager::mouseButtonPress(UInt16 button, Int16 x, Int16 y)
{
	switch (button)
	{
	}

	_mousebuttons |= 1 << button;
	_lastx = x;
	_lasty = y;
}


/*! call when a mouse button is released. button is the number of the pressed
	button, starting at 0, ordered from left to right. A wheel should be
	mapped to buttons 3 and 4. The position is in pixel, starting at the
	upper left corner.
 */
void CAVESceneManager::mouseButtonRelease(UInt16 button, Int16 x, Int16 y)
{
	switch (button)
	{
	}
	_mousebuttons &= ~(1 << button);
	_lastx = x;
	_lasty = y;
}

void CAVESceneManager::key(UChar8 key, Int16 x, Int16 y)
{
	switch ( key )
	{
	}
	// dummy lines to temporarily remove build error
	int a=x;a=y;a++;
}

/*! Calculate a ray that starts at the eye and goes through the position on the
	screen given by x,y.
 */

Line CAVESceneManager::calcViewRay(Int16 x, Int16 y)
{
	Line l;

	_cameras[0]->calcViewRay( l, x, y, *_win->getPort()[0] );

	return l;
}

/*------------------------------ access -----------------------------------*/

/*---------------------------- properties ---------------------------------*/

/*-------------------------- your_category---------------------------------*/

/*-------------------------- assignment -----------------------------------*/

/*-------------------------- comparison -----------------------------------*/

/*! less
 */

bool CAVESceneManager::operator < (const CAVESceneManager &other) const
{
	return this < &other;
}


/*------------------------ internal funcs ---------------------------------*/


//amh - mono camera
ViewportPtr CAVESceneManager::CreateMonoProjCamVeiwport(
		PerspectiveCameraPtr camera, Pnt4f size, ImageBackgroundPtr bg,
		NodePtr user, Pnt3f *vpproj)
{
	// first create the decorator
	ProjectionCameraDecoratorPtr camDec;
	camDec=OSG::ProjectionCameraDecorator::create();
	beginEditCP(camDec);
		camDec->setLeftEye(true);
		camDec->setEyeSeparation(0);
        camDec->setDecoratee(camera);
        camDec->setUser(user);
        camDec->getSurface().push_back(vpproj[0] * _scaleFactor);
        camDec->getSurface().push_back(vpproj[1] * _scaleFactor);
        camDec->getSurface().push_back(vpproj[2] * _scaleFactor);
        camDec->getSurface().push_back(vpproj[3] * _scaleFactor);
    endEditCP(camDec);

    // then the viewport
    ViewportPtr vp = Viewport::create();
    beginEditCP(vp);
        vp->setSize(size[0],size[1],size[2],size[3]);
        vp->setCamera(camDec);
        vp->setBackground(bg);
        vp->setRoot(_internalRoot);
        vp->getForegrounds().push_back(_foreground);
    endEditCP  (vp);
    return vp;
}


//amh - my try at gettign projection stereo going...
StereoBufferViewportPtr CAVESceneManager::CreateProjStereoVeiwport(bool left,
            PerspectiveCameraPtr camera, Pnt4f size, ImageBackgroundPtr bg,
            NodePtr user, Pnt3f *vpproj)
{
    // first create the decorator
    ProjectionCameraDecoratorPtr CamDec;
    CamDec=OSG::ProjectionCameraDecorator::create();
    beginEditCP(CamDec);
        CamDec->setLeftEye(left);
        CamDec->setEyeSeparation(_eyeseparation * _scaleFactor);
        CamDec->setDecoratee(camera);
        CamDec->setUser(user);
        CamDec->getSurface().push_back(vpproj[0] * _scaleFactor);
        CamDec->getSurface().push_back(vpproj[1] * _scaleFactor);
        CamDec->getSurface().push_back(vpproj[2] * _scaleFactor);
        CamDec->getSurface().push_back(vpproj[3] * _scaleFactor);
    endEditCP(CamDec);

    // then the viewport
    StereoBufferViewportPtr vp = StereoBufferViewport::create();
	beginEditCP(vp);
		vp->setSize(size[0],size[1],size[2],size[3]);
		vp->setCamera(CamDec);
		vp->setBackground(bg);
		vp->setRoot(_internalRoot);
		vp->setLeftBuffer(left);
		vp->setRightBuffer(!left);
		vp->getForegrounds().push_back(_foreground);
	endEditCP  (vp);


    return vp;
}


void CAVESceneManager::CreateWand(NodePtr parentNode)
{
    // exit if we already have a wand
    if (_wand != NullFC){
        return;
    }

    // default wand is a cylinder
    _wand = makeCylinder(20, 2, 8, true, true, true);
    GeometryPtr geo = GeometryPtr::dcast(_wand->getCore());

    SimpleMaterialPtr sm = SimpleMaterial::create();
    beginEditCP(sm, SimpleMaterial::DiffuseFieldMask |
                    SimpleMaterial::LitFieldMask);
    {
        sm->setDiffuse(Color3f( 1, 1, 0));
    }
    endEditCP  (sm, SimpleMaterial::DiffuseFieldMask |
                    SimpleMaterial::LitFieldMask);

    beginEditCP(geo, Geometry::MaterialFieldMask);
    {
        geo->setMaterial(sm);
    }
    endEditCP  (geo, Geometry::MaterialFieldMask);

    // create a transformation for the wand
    _wandTransform = ComponentTransform::create();
    // temporily set the transform
    beginEditCP(_wandTransform, ComponentTransform::TranslationFieldMask);
        _wandTransform->setTranslation(Vec3f(0,0,0));
    endEditCP(_wandTransform, ComponentTransform::TranslationFieldMask);

    _wandTransformNode = Node::create();
    beginEditCP(_wandTransformNode, Node::ChildrenFieldMask | Node::CoreFieldMask);
        _wandTransformNode->addChild(_wand);
        _wandTransformNode->setCore(_wandTransform);
    endEditCP(_wandTransformNode, Node::ChildrenFieldMask | Node::CoreFieldMask);

    // and add it to the parentNode
    beginEditCP(parentNode, Node::ChildrenFieldMask );
        parentNode->addChild(_wandTransformNode);
    endEditCP(parentNode, Node::ChildrenFieldMask );

}

//Shadow support
void CAVESceneManager::enableShadows( void )
{
    _useShadows = true;
}

void CAVESceneManager::addShadowLightSource(NodePtr light)
{
	if(_useShadows)
	{
		for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++)
		{
			ShadowViewportPtr svp = ShadowViewportPtr::dcast(_win->getPort(portCnt));
			beginEditCP(svp);
				svp->getLightNodes().push_back(light);
			endEditCP(svp);
		}
	}
}

void CAVESceneManager::removeShadowLightSource(NodePtr light)
{
	if(_useShadows)
	{
		for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++)
		{
			ShadowViewportPtr svp = ShadowViewportPtr::dcast(_win->getPort(portCnt));
			beginEditCP(svp);
				svp->getLightNodes().erase(svp->getLightNodes().find(light));
			endEditCP(svp);
		}
	}
}

void CAVESceneManager::addExclusionFromShadow(NodePtr exclude)
{
	if (_useShadows)
	{
		for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++)
		{
			ShadowViewportPtr svp = ShadowViewportPtr::dcast(_win->getPort(portCnt));
			beginEditCP(svp);
				svp->getExcludeNodes().push_back(exclude);
			endEditCP(svp);
		}
	}
}

void CAVESceneManager::removeExclusionFromShadow(NodePtr include)
{
	if (_useShadows)
	{
		for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++)
		{
			ShadowViewportPtr svp = ShadowViewportPtr::dcast(_win->getPort(portCnt));
			beginEditCP(svp);
				svp->getExcludeNodes().erase(svp->getExcludeNodes().find(include));
			endEditCP(svp);
		}
	}
}

ShadowViewportPtr CAVESceneManager::CreateMonoShadowViewport(
        PerspectiveCameraPtr camera, Pnt4f size, ImageBackgroundPtr bg,
        NodePtr user, Pnt3f *vpproj)
{
    // first create the decorator
    ProjectionCameraDecoratorPtr camDec;
    camDec=OSG::ProjectionCameraDecorator::create();
    beginEditCP(camDec);
        camDec->setLeftEye(true);
        camDec->setEyeSeparation(0);
        camDec->setDecoratee(camera);
        camDec->setUser(user);
        camDec->getSurface().push_back(vpproj[0] * _scaleFactor);
        camDec->getSurface().push_back(vpproj[1] * _scaleFactor);
        camDec->getSurface().push_back(vpproj[2] * _scaleFactor);
        camDec->getSurface().push_back(vpproj[3] * _scaleFactor);
    endEditCP(camDec);

    // then the viewport

    printf("\n\nViewport with Shadow support created!\n\n");

    ShadowViewportPtr svp = ShadowViewport::create();
    beginEditCP(svp);
        svp->setSize(size[0],size[1],size[2],size[3]);
        svp->setCamera(camDec);
        svp->setBackground(bg);
        svp->setRoot(_internalRoot);
        svp->getForegrounds().push_back(_foreground);

	// shadow
	svp->setOffFactor(4.0);
	svp->setOffBias(8.0);
	svp->setGlobalShadowIntensity(0.9);
	svp->setMapSize(1024);
	svp->setShadowMode(ShadowViewport::STD_SHADOW_MAP);
	//svp->setShadowColor(Color4f(0.1,0.1,0.1,1.0));
	svp->setShadowSmoothness(0.0);
    endEditCP  (svp);

    return svp;
}

ShadowViewportPtr CAVESceneManager::CreateStereoShadowViewport(bool left,
            PerspectiveCameraPtr camera, Pnt4f size, ImageBackgroundPtr bg,
            NodePtr user, Pnt3f *vpproj)
{
    // first create the decorator
    ProjectionCameraDecoratorPtr CamDec;
    CamDec=OSG::ProjectionCameraDecorator::create();
    beginEditCP(CamDec);
        CamDec->setLeftEye(left);
        CamDec->setEyeSeparation(_eyeseparation * _scaleFactor);
        CamDec->setDecoratee(camera);
        CamDec->setUser(user);
        CamDec->getSurface().push_back(vpproj[0] * _scaleFactor);
        CamDec->getSurface().push_back(vpproj[1] * _scaleFactor);
        CamDec->getSurface().push_back(vpproj[2] * _scaleFactor);
        CamDec->getSurface().push_back(vpproj[3] * _scaleFactor);
    endEditCP(CamDec);

    // then the viewport
    //modification for shadow support

    printf("\n\nViewport with Shadow support created!\n\n");

    	ShadowViewportPtr svp = ShadowViewport::create();
	beginEditCP(svp);
		svp->setSize(size[0],size[1],size[2],size[3]);
		svp->setCamera(CamDec);
		svp->setBackground(bg);
		svp->setRoot(_internalRoot);
		svp->setLeftBuffer(left);
		svp->setRightBuffer(!left);
		svp->getForegrounds().push_back(_foreground);

		//shadow
		svp->setOffFactor(4.0);
		svp->setOffBias(8.0);
		svp->setGlobalShadowIntensity(0.9);
		svp->setMapSize(1024);
		svp->setShadowMode(ShadowViewport::STD_SHADOW_MAP);
		//svp->setShadowColor(Color4f(0.1,0.1,0.1,1.0));
		svp->setShadowSmoothness(0.0);
	endEditCP  (svp);



    return svp;
}

void CAVESceneManager::setShadowOffBias(const Real32 &value)
{
	if (_useShadows)
	{
		for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++)
		{
			ShadowViewportPtr svp = ShadowViewportPtr::dcast(_win->getPort(portCnt));
			beginEditCP(svp);
				svp->setOffBias(value);
			endEditCP(svp);
		}
	}
}
void CAVESceneManager::setShadowOffFactor(const Real32 &value)
{
	if (_useShadows)
	{
		for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++)
		{
			ShadowViewportPtr svp = ShadowViewportPtr::dcast(_win->getPort(portCnt));
			beginEditCP(svp);
				svp->setOffFactor(value);
			endEditCP(svp);
		}
	}
}
void CAVESceneManager::setShadowMapSize(const UInt32 &value)
{
	if (_useShadows)
	{
		for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++)
		{
			ShadowViewportPtr svp = ShadowViewportPtr::dcast(_win->getPort(portCnt));
			beginEditCP(svp);
				svp->setMapSize(value);
			endEditCP(svp);
		}
	}
}
void CAVESceneManager::setShadowMode(const UInt32 &value)
{
	if (_useShadows)
	{
		for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++)
		{
			ShadowViewportPtr svp = ShadowViewportPtr::dcast(_win->getPort(portCnt));
			beginEditCP(svp);
				svp->setShadowMode(value);
			endEditCP(svp);
		}
	}
}
void CAVESceneManager::setShadowSmoothness(const Real32 &value)
{
	if (_useShadows)
	{
		for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++)
		{
			ShadowViewportPtr svp = ShadowViewportPtr::dcast(_win->getPort(portCnt));
			beginEditCP(svp);
				svp->setShadowSmoothness(value);
			endEditCP(svp);
		}
	}
}
void CAVESceneManager::setShadowOn(const bool &value)
{
	if (_useShadows)
	{
		for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++)
		{
			ShadowViewportPtr svp = ShadowViewportPtr::dcast(_win->getPort(portCnt));
			beginEditCP(svp);
				svp->setShadowOn(value);
			endEditCP(svp);
		}
	}
}
void CAVESceneManager::setShadowAutoSearchForLights(const bool &value)
{
	if (_useShadows)
	{
		for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++)
		{
			ShadowViewportPtr svp = ShadowViewportPtr::dcast(_win->getPort(portCnt));
			beginEditCP(svp);
				svp->setAutoSearchForLights(value);
			endEditCP(svp);
		}
	}
}
void CAVESceneManager::setShadowGlobalShadowIntensity(const Real32 &value)
{
	if (_useShadows)
	{
		for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++)
		{
			ShadowViewportPtr svp = ShadowViewportPtr::dcast(_win->getPort(portCnt));
			beginEditCP(svp);
				svp->setGlobalShadowIntensity(value);
			endEditCP(svp);
		}
	}
}
void CAVESceneManager::setShadowAutoExcludeTransparentNodes(const bool &value)
{
	if (_useShadows)
	{
		for (int portCnt=0; portCnt<_win->getPort().size(); portCnt++)
		{
			ShadowViewportPtr svp = ShadowViewportPtr::dcast(_win->getPort(portCnt));
			beginEditCP(svp);
				svp->setAutoExcludeTransparentNodes(value);
			endEditCP(svp);
		}
	}
}



/*------------------------------------------------------------------------*/
/*                              cvs id's                                  */

#ifdef OSG_SGI_CC
#pragma set woff 1174
#endif

#ifdef OSG_LINUX_ICC
#pragma warning( disable : 177 )
#endif

namespace
{
    static Char8 cvsid_cpp[] = "@(#)$Id: OSGCAVESceneManager.cpp,v 1.1.1.1 2004/10/07 01:01:10 adrian Exp $";
    static Char8 cvsid_hpp[] = OSGCAVESCENEMANAGER_HEADER_CVSID;
//    static Char8 cvsid_inl[] = OSGCAVESCENEMANAGER_INLINE_CVSID;
}
