/*---------------------------------------------------------------------------*\
 *           interactive networked Virtual Reality system (inVRs)            *
 *                                                                           *
 *    Copyright (C) 2005-2009 by the Johannes Kepler University, Linz        *
 *                                                                           *
 *                            www.inVRs.org                                  *
 *                                                                           *
 *              contact: canthes@inVRs.org, rlander@inVRs.org                *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                        Project: CAVESceneManager                          *
 *                                                                           *
 * The CAVE Scene Manager (CSM) is a wrapper for the OpenSG multi-display    *
 * functionality, which allows the easy configuration of a multi-display     *
 * setup via text files. It was developed by Adrian Haffegee of the Centre   *
 * for Advanced Computing and Emerging Technologies (ACET) at the            *
 * University of Reading, UK as part of his MSc Thesis.                      *
 *                                                                           *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                Changes                                    *
 *                                                                           *
 *                                                                           *
\*---------------------------------------------------------------------------*/

/* \file OSGCAVEWall.cpp
 *
 *  Adrian Haffegee, Sept 2004
 *	Implementation file for CAVE Display wall class
 */
#include <stdlib.h>
#include <stdio.h>

#include <OpenSG/OSGBaseFunctions.h>

#include "../include/OpenSG/OSGCAVEWall.h"


OSG_USING_NAMESPACE



/**
 *  Constructor
 * @param name name of the new wall to be created
 */
CAVEWall::CAVEWall(char *name)
{
	m_active = false;
	m_name = strdup(name);
	m_window = false;
	m_pipe = NULL;
	m_geometry = NULL;
	/*    printf("\nWallCfg1: new wall len:%d:%d\n", strlen(name), strlen(m_name));
		  printf("\nWallCfg2: new wall<%s>\n", name);
		  printf("\nWallCfg3: new wall[%s]\n", m_name);
		  printf("\nWallCfg4: new wall\n");*/
	if (strcmp(name, "front")==0) {
		m_wallPos = wallFront;
	} else if (strcmp(name, "left")==0) {
		m_wallPos = wallLeft;
	} else if (strcmp(name, "right")==0) {
		m_wallPos = wallRight;
	} else if (strcmp(name, "floor")==0) {
		m_wallPos = wallFloor;
	} else if (strcmp(name, "back")==0) {
		m_wallPos = wallBack;
	} else if (strcmp(name, "ceiling")==0) {
		m_wallPos = wallCeiling;
	} else {
		m_wallPos = wallUserDefined;
	}
	for (int i=0; i<4; i++) {
		m_projData[i] = Pnt3f(0,0,0);
	}
	m_pipeIndex=-1;
}


/**
 *  Destructor
 */
CAVEWall::~CAVEWall(void)
{
	free(m_name);
	if (m_pipe) {
		free(m_pipe);
	}
}


/**
 *  display wall configuration
 */
void CAVEWall::printConfig()
{
	printf("\tCAVE Wall:%s ", getName());
	if (isActive()) {
		printf("- active, pipe '%s' (index:%d)%s", getPipe(), getPipeIndex(),
				getWindow() ? ", windowed" : "");
		if (getGeometry()) {
			printf(", geometry '%dx%d+%d+%d'", m_geomCoordMaxX-m_geomCoordMinX,
					m_geomCoordMaxY-m_geomCoordMinY, m_geomCoordMinX, m_geomCoordMinY );
		}
		printf("\n\t\tProj data:");
		for (int i=0; i<4; i++) {
			printf(" %.3f,%.3f,%.3f ", m_projData[i][0], m_projData[i][1], m_projData[i][2] );
		}
		printf("\n");
	} else {
		printf("- not active\n");
	}
	//    printf("\n\tDisplay Mode: %s", isStereo()?"stereo":"mono");
}


/** checks + sets projection data for walls
 * @param origin for the CAVE
 * @param width width of CAVE
 * @param height height of CAVE
 * @return true if projrction data set ok
 */
bool CAVEWall::checkProjectionData(Pnt3f origin, float width, float height)
{

	// check we have cave width/height (both must be non 0 for a CAVE display)
	switch(m_wallPos){
		case wallFront:
		case wallLeft:
		case wallRight:
		case wallFloor:
		case wallBack:
		case wallCeiling:
			if ((height == 0) || (width == 0)) {
#ifdef CSM_DEBUG
				printf("Cfg: CAVE height/width not configured\n");
#endif
				return false;
			}
			break;
		default:
			break;
	}
	// if this is a regular cave wall, projData is calculated here
	//    printf("\nWallCfg: wall pos :%d (%.3f, %.3f, %.3f, %.3f, %.3f)\n", m_wallPos,
	//            origin[0], origin[1], origin[0], width, height);
	switch(m_wallPos){
		case wallFront:
			m_projData[0] = Pnt3f(0-origin[0], 0-origin[1], 0-(width-origin[2]));
			m_projData[1] = Pnt3f(0+(width-origin[0]), 0-origin[1], 0-(width-origin[2]));
			m_projData[2] = Pnt3f(0+(width-origin[0]), 0+(height-origin[1]), 0-(width-origin[2]));
			m_projData[3] = Pnt3f(0-origin[0], 0+(height-origin[1]), 0-(width-origin[2]));
			break;

		case wallLeft:
			m_projData[0] = Pnt3f(0-origin[0], 0-origin[1], 0+origin[2]);
			m_projData[1] = Pnt3f(0-origin[0], 0-origin[1], 0-(width-origin[2]));
			m_projData[2] = Pnt3f(0-origin[0], 0+(height-origin[1]), 0-(width-origin[2]));
			m_projData[3] = Pnt3f(0-origin[0], 0+(height-origin[1]), 0+origin[2]);
			break;

		case wallRight:
			m_projData[0] = Pnt3f(0+(width-origin[0]), 0-origin[1], 0-(width-origin[2]));
			m_projData[1] = Pnt3f(0+(width-origin[0]), 0-origin[1], 0+origin[2]);
			m_projData[2] = Pnt3f(0+(width-origin[0]), 0+(height-origin[1]), 0+origin[2]);
			m_projData[3] = Pnt3f(0+(width-origin[0]), 0+(height-origin[1]), 0-(width-origin[2]));
			break;

		case wallFloor:
			m_projData[0] = Pnt3f(0-origin[0], 0-origin[1], 0+origin[2]);
			m_projData[1] = Pnt3f(0+(width-origin[0]), 0-origin[1], 0+origin[2]);
			m_projData[2] = Pnt3f(0+(width-origin[0]), 0-origin[1], 0-(width-origin[2]));
			m_projData[3] = Pnt3f(0-origin[0], 0-origin[1], 0-(width-origin[2]));
			break;
			// not too sure about these next two, since we can't check them....
		case wallBack:
			m_projData[0] = Pnt3f(0+(width-origin[0]), 0-origin[1], 0+origin[2]);
			m_projData[1] = Pnt3f(0-origin[0], 0-origin[1], 0+origin[2]);
			m_projData[2] = Pnt3f(0-origin[0], 0+(height-origin[1]), 0+origin[2]);
			m_projData[3] = Pnt3f(0+(width-origin[0]), 0+(height-origin[1]), 0+origin[2]);
			break;

		case wallCeiling:
			m_projData[0] = Pnt3f(0-origin[0], 0-origin[1], 0+origin[2]);
			m_projData[1] = Pnt3f(0+(width-origin[0]), 0-origin[1], 0+origin[2]);
			m_projData[2] = Pnt3f(0+(width-origin[0]), 0-origin[1], 0-(width-origin[2]));
			m_projData[3] = Pnt3f(0-origin[0], 0-origin[1], 0-(width-origin[2]));

			/*
			   m_projData[0] = Pnt3f(0-origin[0], 0+(height-origin[1]), 0-(width-origin[2]));
			   m_projData[1] = Pnt3f(0+(width-origin[0]), 0+(height-origin[1]), 0-(width-origin[2]));
			   m_projData[2] = Pnt3f(0+(width-origin[0]), 0+(height-origin[1]), 0+origin[2]);
			   m_projData[3] = Pnt3f(0-origin[0], 0+(height-origin[1]), 0+origin[2]);
			 */
			break;

		default:
			// otherwise just check there is some data
			if ((m_projData[0] == Pnt3f(0,0,0)) &&
					(m_projData[1] == Pnt3f(0,0,0)) &&
					(m_projData[2] == Pnt3f(0,0,0)) &&
					(m_projData[3] == Pnt3f(0,0,0)) ) {
				return false;
			}
			return true;
	}
	return true;
}

/*! set active
 * @param activate wheter to set active or not
 */
void CAVEWall::setActive(bool activate)
{
	m_active = activate;

}

/*! set gfx pipe
 * @param pipe string of pipe for wall
 * @return true if set ok
 */
bool CAVEWall::setPipe(char *pipe)
{
	if (!pipe) {
		return false;
	}
	if (m_pipe) {
		free(m_pipe);
	}
	// if the pipe value passed in is just a single number it should
	// be prefixed with ":0."
	if ((pipe[0] == '0') || (atoi(pipe) > 0)) {
		char prefix[] = ":0.";
		if (m_pipe = (char*)malloc(strlen(pipe)+strlen(prefix)+1)) {
			strcpy(m_pipe, prefix);
			strcat(m_pipe, pipe);
		}
	} else {
		m_pipe = strdup(pipe);
	}
#ifdef CSM_DEBUG
	printf("Configuring pipe with '%s'\n", m_pipe);
#endif
	// could do some checking???
	return true;
}

/*! set if window or full screen
 * @param window if window
 */
void CAVEWall::setWindow(bool window)
{
	m_window = window;
}

/*! set display geometry
 * @param geometry string of containing geometry info
 * @return true if set ok
 */
bool CAVEWall::setGeometry(char *geometry)
{
	if (!geometry) {
		return false;
	}
	if (m_geometry) {
		free(m_geometry);
	}

	// check format of geometry (XDIMxYDIM+XOFFSET+YOFFSET)
	int dimX, dimY;
	char* charPos;
	if (charPos = strchr(geometry, 'x')) {
		charPos++;
	}
	if ((charPos) && (dimX=atoi(geometry)) && (dimY=atoi(charPos))){

		if (charPos = strchr(charPos, '+')) {
			charPos++;
			m_geomCoordMinX = atoi(charPos);

			if (charPos = strchr(charPos, '+')) {
				charPos++;
				m_geomCoordMinY = atoi(charPos);
				m_geomCoordMaxX = m_geomCoordMinX + dimX;
				m_geomCoordMaxY = m_geomCoordMinY + dimY;

#ifdef CSM_DEBUG
				printf("Configuring geo with '%s'\n", geometry);
#endif
				m_geometry = strdup(geometry);
				return true;
			}
		}
	}
	printf("Invalid geometry config '%s'\n", geometry);
	m_geometry=NULL;
	return false;
}

/*! set display index
 * @param index index to use for this wall
 */
void CAVEWall::setPipeIndex(int index)
{
	m_pipeIndex = index;
}

/*! set projection data
 * @param lowerLeft coordinates of lower left corner of screen
 * @param upperLeft coordinates of upper left corner of screen
 * @param lowerRight coordinates of lower right corner of screen
 * @param convFact conversion factor to convert units (into cm)
 */
void CAVEWall::setProjectionData(Pnt3f lowerLeft, Pnt3f upperLeft, Pnt3f lowerRight, float convFact)
{
	m_projData[0] = Pnt3f(lowerLeft[0]*convFact, lowerLeft[1]*convFact, lowerLeft[2]*convFact);
	m_projData[1] = Pnt3f(lowerRight[0]*convFact, lowerRight[1]*convFact, lowerRight[2]*convFact);
	// upper right can be constructed from the other ones...
	m_projData[2] = Pnt3f((upperLeft[0]+ lowerRight[0] - lowerLeft[0]) * convFact,
			(upperLeft[1]+ lowerRight[1] - lowerLeft[1]) * convFact,
			(upperLeft[2]+ lowerRight[2] - lowerLeft[2]) * convFact);
	m_projData[3] = Pnt3f(upperLeft[0]*convFact, upperLeft[1]*convFact, upperLeft[2]*convFact);
}

/*! get name
 * @return name of wall
 */
char* CAVEWall::getName(void)
{
	return m_name;
}

/*! if wall active
 * @return if wall active
 */
bool CAVEWall::isActive(void)
{
	return m_active;
}

/*! get pipe
 * @return string of wall pipe
 */
char* CAVEWall::getPipe(void)
{
	return m_pipe;
}

/*! get window
 * @return if a window
 */
bool CAVEWall::getWindow(void)
{
	return m_window;
}

/*! get geometry
 * @return string of geometry
 */
char* CAVEWall::getGeometry(void)
{
	return m_geometry;
}

/*! get geometry minimum X geometry of wall
 * @return min X geometry
 */
int CAVEWall::getGeomCoordMinX(void)
{
	return m_geomCoordMinX;
}

/*! get geometry minimum Y geometry of wall
 * @return min Y geometry
 */
int CAVEWall::getGeomCoordMaxX(void)
{
	return m_geomCoordMaxX;
}

/*! get geometry maximum X geometry of wall
 * @return max X geometry
 */
int CAVEWall::getGeomCoordMinY(void)
{
	return m_geomCoordMinY;
}

/*! get geometry maximum Y geometry of wall
 * @return max Y geometry
 */
int CAVEWall::getGeomCoordMaxY(void)
{
	return m_geomCoordMaxY;
}

/*! get wall projection data
 * @return array of points for projrction data
 */
Pnt3f* CAVEWall::getProjData(void)
{
	return m_projData;
}

/*! get pipe index
 * @return pipe index
 */
int CAVEWall::getPipeIndex(void)
{
	return m_pipeIndex;
}






