/*---------------------------------------------------------------------------*\
 *           interactive networked Virtual Reality system (inVRs)            *
 *                                                                           *
 *    Copyright (C) 2005-2009 by the Johannes Kepler University, Linz        *
 *                                                                           *
 *                            www.inVRs.org                                  *
 *                                                                           *
 *              contact: canthes@inVRs.org, rlander@inVRs.org                *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                                License                                    *
 *                                                                           *
 * This library is free software; you can redistribute it and/or modify it   *
 * under the terms of the GNU Library General Public License as published    *
 * by the Free Software Foundation, version 2.                               *
 *                                                                           *
 * This library is distributed in the hope that it will be useful, but       *
 * WITHOUT ANY WARRANTY; without even the implied warranty of                *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 * Library General Public License for more details.                          *
 *                                                                           *
 * You should have received a copy of the GNU Library General Public         *
 * License along with this library; if not, write to the Free Software       *
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                 *
\*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*\
 *                        Project: CAVESceneManager                          *
 *                                                                           *
 * The CAVE Scene Manager (CSM) is a wrapper for the OpenSG multi-display    *
 * functionality, which allows the easy configuration of a multi-display     *
 * setup via text files. It was developed by Adrian Haffegee of the Centre   *
 * for Advanced Computing and Emerging Technologies (ACET) at the            *
 * University of Reading, UK as part of his MSc Thesis.                      *
 *                                                                           *
\*---------------------------------------------------------------------------*/

/*! \file appctrl.cpp
 *
 *  Adrian Haffegee, June 2005
 *  implementation for the application control utilities
 */

//#include <stdlib.h>
//#include <stdio.h>
//#include <string.h>

#include "../include/OpenSG/appctrl.h"

OSG_USING_NAMESPACE

#if defined(WIN32)
#include <Windows.h>
#endif

/*! \brief structure ...
 */
struct tmpStruct{
	int a;
};

#ifdef WIN32

/*! start server applications for rendering the display
 *
 * @param cfg CAVE configuration file to use for application
 * @return 0 if started ok, <0 if servers failed
 */
int startServers(CAVEConfig *cfg, char* serverAdress, char* serverPath)
{
	// work out path to servers
	const char* serveType = (cfg->isStereo() ? "server-stereo" : "server-mono");
	char servePth[255];
	if (serverPath)
		_snprintf(servePth, 255, "%s\\%s", serverPath, serveType);
	else
		_snprintf(servePth, 255, "%s", serveType);
	fprintf(stderr, "\nOpening servers from %s\n", servePth);
	char cmdLine[2048];
	char cmdLineBuffer[2048];

	if(serverAdress == NULL) serverAdress = "localhost";

	csmServerHandles = new PROCESS_INFORMATION[cfg->getNumActiveWalls()];
	// spawn a process for each of the servers - 1 for each active wall (that doesn't have a unique pipe)
	for (int wallCnt=0; wallCnt<cfg->getNumActiveWalls(); wallCnt++) {
		// if one of the previous walls used this pipe, continue to next wall
		bool pipeUsed=false;
		for (int prevWall=0; prevWall<wallCnt; prevWall++) {
			if (cfg->getActiveWall(wallCnt)->getPipeIndex() ==
					cfg->getActiveWall(prevWall)->getPipeIndex()) {
				pipeUsed=true;
			}
		}
		if (pipeUsed) {
			continue;
		}

		CAVEWall* wall = cfg->getActiveWall(wallCnt);
		char wallName[256];
		strcpy(wallName, wall->getName());
		int rc;

		sprintf(cmdLineBuffer, "%s %s -geometry 500x500+20+20 -display %s -a %s ",
			 servePth, serveType, wall->getPipe(), serverAdress);
		if (wall->getWindow())
		{
			sprintf(cmdLine, "-w %s", wallName);
		} else {
			sprintf(cmdLine, "%s", wallName);
		}
		strcat(cmdLineBuffer, cmdLine);
		fprintf(stderr, "startServers(): executing '%s %s'\n", servePth, cmdLineBuffer);

		if(!strcmp(serverAdress, "127.0.0.1") || !strcmp(serverAdress, "localhost"))
		{
			STARTUPINFO si;
			ZeroMemory( &si, sizeof(si) );
			si.cb = sizeof(si);
			ZeroMemory( &csmServerHandles[wallCnt], sizeof(csmServerHandles[wallCnt]) );

			if (!CreateProcess(NULL, cmdLineBuffer, NULL, NULL, FALSE, 0, NULL, NULL, &si, &(csmServerHandles[wallCnt]))) {
				fprintf(stderr, "startServers(): error at starting server! Please start server manually!\n");
				csmServerHandles[wallCnt].hProcess = INVALID_HANDLE_VALUE;
				csmServerHandles[wallCnt].hThread = INVALID_HANDLE_VALUE;
			} // if
			else {
				assert(csmServerHandles[wallCnt].hProcess != INVALID_HANDLE_VALUE);
			} // else
//			ShellExecute(NULL, NULL, servePth, cmdLineBuffer, NULL, SW_SHOW);
		} else
		{
			fprintf(stderr, "startServers(): please start server at %s\n", serverAdress);
		} // else
	}
	Sleep(1000);
	return 0;
}


// Create a window for the app to write to
MultiDisplayWindowPtr createAppWindow(CAVEConfig *cfg)
{
	// Create a window
	fprintf(stderr, "createAppWindow(): entering method\n");
	MultiDisplayWindowPtr mwin= MultiDisplayWindow::create();
	// configure it, + make a connection between it + the servers (renderers)
	beginEditCP(mwin);
	// multicast if required
	if (cfg->isMCastServer()) {
		mwin->setConnectionType("Multicast");
	}
	// dummy size for navigator
	//amh - what does this do?
	mwin->setSize(1136*4,1136);
	//    mwin->setSize(1136,1136);
	// push back servers
	for (int wallCnt=0; wallCnt<cfg->getNumActiveWalls(); wallCnt++) {
		// if one of the previous walls has same pipe index, continue to next wall
		bool pipeUsed=false;
		for (int prevWall=0; prevWall<wallCnt; prevWall++) {
			if (cfg->getActiveWall(wallCnt)->getPipeIndex() ==
					cfg->getActiveWall(prevWall)->getPipeIndex()) {
				pipeUsed=true;
			}
		}
		if (pipeUsed) {
			continue;
		}
		// add set this wall name as one of the servers
		char* server = cfg->getActiveWall(wallCnt)->getName();
		fprintf(stderr, "\nAdding server: %s\n",server);
		mwin->getServers().push_back(server);
	}
	endEditCP(mwin);
	fprintf(stderr,"createAppWindow(): leaving method\n");
	return mwin;
}
#else
/*! start server applications for rendering the display
 *
 * @param cfg CAVE configuration file to use for application
 * @return 0 if started ok, <0 if servers failed
 */
int startServers(CAVEConfig *cfg, char* serverAdress, char* serverPath)
{
	// work out path to servers
	const char* serveType = (cfg->isStereo() ? "server-stereo" : "server-mono");
	char servePth[255];
	snprintf(servePth, 255, "%s/%s", serverPath?serverPath:".",serveType);
	fprintf(stderr, "\nOpening servers from %s\n", servePth);
	char cmdLine[2048];
	char cmdLineBuffer[2048];

	if(serverAdress == NULL) serverAdress = "localhost";

	// spawn a process for each of the servers - 1 for each active wall (that doesn't have a unique pipe)
	for (int wallCnt=0; wallCnt<cfg->getNumActiveWalls(); wallCnt++) {
		// if one of the previous walls used this pipe, continue to next wall
		bool pipeUsed=false;
		for (int prevWall=0; prevWall<wallCnt; prevWall++) {
			if (cfg->getActiveWall(wallCnt)->getPipeIndex() ==
					cfg->getActiveWall(prevWall)->getPipeIndex()) {
				pipeUsed=true;
			}
		}
		if (pipeUsed) {
			continue;
		}

		switch ( fork()) {
			case -1:
				perror("fork() failed");
				return -1;
				break;

			case 0:
				{
					CAVEWall* wall = cfg->getActiveWall(wallCnt);
					char wallName[256];
					strcpy(wallName, wall->getName());
					int rc;

					if (cfg->isStereo())
					{
						sprintf(cmdLineBuffer, "%s %s -geometry 500x500+20+20 -display %s -a %s ",
							 servePth, serveType, wall->getPipe(), serverAdress);
						if (wall->getWindow())
						{
							sprintf(cmdLine, "-w %s", wallName);
						} else {
							sprintf(cmdLine, "%s", wallName);
						}
						strcat(cmdLineBuffer, cmdLine);
						fprintf(stderr, "startServers(): executing '%s'\n", cmdLineBuffer);

						if(!strcmp(serverAdress, "127.0.0.1") || !strcmp(serverAdress, "localhost"))
						{
							rc = execl(servePth, serveType, "-geometry", "500x500+20+20",
								 "-display", wall->getPipe(), "-a", "localhost",
							 	 wall->getWindow()?"-w":"", wallName, (void*)NULL);
						} else
						{
							fprintf(stderr, "startServers(): please start server at %s\n", serverAdress);
						}
					} else {
						if (strcmp(wall->getPipe(),":0.0") != 0 )
						{
							printf("Starting Server: %s %s -geometry 500x500+20+20 -display %s -a localhost %s %s\n",
								servePth, serveType, wall->getPipe(),  wall->getWindow()?"-w":"", wallName);
							rc = execl(servePth, serveType, "-geometry", "500x500+20+20",
									"-display", wall->getPipe(), "-a", "localhost",
									wall->getWindow() ? "-w" : "", wallName, 0);
						} else {
							sprintf(cmdLineBuffer, "%s %s -geometry 500x500+20+20 -a %s ", servePth, serveType, serverAdress);
							if (wall->getWindow())
							{
								sprintf(cmdLine, "-w %s", wall->getName());
							} else {
								sprintf(cmdLine, "%s", wall->getName());
							}
							strcat(cmdLineBuffer, cmdLine);
							fprintf(stderr, "startServers(): executing '%s'\n", cmdLineBuffer);

							if (!strcmp(serverAdress, "127.0.0.1") || !strcmp(serverAdress, "localhost"))
							{
								rc = execl(servePth, serveType, "-geometry", "500x500+20+20", "-a", "localhost",
									wall->getWindow() ? "-w":"", wallName, (void*)NULL);
							} else {
								fprintf(stderr, "startServers(): please start server at %s\n", serverAdress);
							}
						}
					}
					if (rc == -1) {
						perror("execl failed!");
					}
					// never get here if execl success, since control of process would have moved to exec
					exit(0);
					return -2;
					break;
				}
			default:
				fprintf(stderr, "continuing with parent process\n");
				// parent process carries on as normal
				break;
		}
	}
	sleep(1);
	return 0;
}


// Create a window for the app to write to
MultiDisplayWindowPtr createAppWindow(CAVEConfig *cfg)
{
	// Create a window
	fprintf(stderr, "createAppWindow(): entering method\n");
	MultiDisplayWindowPtr mwin= MultiDisplayWindow::create();
	// configure it, + make a connection between it + the servers (renderers)
	beginEditCP(mwin);
	// multicast if required
	if (cfg->isMCastServer()) {
		mwin->setConnectionType("Multicast");
	}
	// dummy size for navigator
	//amh - what does this do?
	mwin->setSize(1136*4,1136);
	//    mwin->setSize(1136,1136);
	// push back servers
	for (int wallCnt=0; wallCnt<cfg->getNumActiveWalls(); wallCnt++) {
		// if one of the previous walls has same pipe index, continue to next wall
		bool pipeUsed=false;
		for (int prevWall=0; prevWall<wallCnt; prevWall++) {
			if (cfg->getActiveWall(wallCnt)->getPipeIndex() ==
					cfg->getActiveWall(prevWall)->getPipeIndex()) {
				pipeUsed=true;
			}
		}
		if (pipeUsed) {
			continue;
		}
		// add set this wall name as one of the servers
		char* server = cfg->getActiveWall(wallCnt)->getName();
		fprintf(stderr, "\nAdding server: %s\n",server);
		mwin->getServers().push_back(server);
	}
	endEditCP(mwin);
	fprintf(stderr,"createAppWindow(): leaving method\n");
	return mwin;
}
#endif // WIN32




